//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include <windows.h>
#include <vd2/system/error.h>
#include <vd2/system/file.h>
#include <vd2/system/registry.h>
#include <vd2/system/strutil.h>
#include <vd2/system/vdstl.h>
#include <vd2/system/w32assist.h>
#include "resource.h"
#include "parser.h"
#include "disasm.h"
#include "log.h"
#include "interp.h"
#include "jit_scalar.h"
#include "ui.h"
#include "editor.h"
#include "texteditor.h"
#include <vd2/plugin/vdvideofiltold.h>

extern "C" IMAGE_DOS_HEADER __ImageBase;

///////////////////////////////////////////////////////////////////////////////
class VDShaderEditor : public VDShaderEditorBaseWindow, public IVDShaderLogOutput, public IVDShaderFilterSource, public IVDTextEditorCallback {
public:
	VDShaderEditor(VDShaderFilterConfig& config, IVDFilterPreview *ifp, IVDFilterPreview2 *ifp2);
	~VDShaderEditor();

	bool Display(HWND hwndParent);

protected:
	void ReadSource(VDStringA& dst);

protected:
	LRESULT WndProc(UINT message, WPARAM wParam, LPARAM lParam);

	void OnCreate();
	void OnDestroy();
	void OnNCDestroy();
	void OnSize();
	bool OnTimer(int id);
	void OnInitMenu();
	bool OnCommand(int cmd);
	void OnFindMsg();

	void WriteLogOutput(const char *s);

	void Compile();
	void UpdateCaption();

	void DoFileNew();
	void DoFileOpen();
	void DoFileOpen(const char *s);
	void DoFileSave();
	void DoFileSaveAs();
	void DoFileExit();
	void DoEditFind();
	void DoEditFindNext();
	void DoEditPrevError();
	void DoEditNextError();
	void DoOptionsChooseFont();
	bool HighlightError(int line);

	bool QuerySaveFirst();
	void OpenPreset(int index);
	void LoadPresets();
	static BOOL CALLBACK StaticLoadPreset(HMODULE hModule, LPCTSTR lpszType, LPTSTR lpszName, LONG_PTR lParam);
	void LoadPreset(HMODULE hModule, LPCTSTR lpszType, LPTSTR lpszName, bool addToMenu);

	void LoadMRUList();
	void SaveMRUList();
	void UpdateMRUListMenu();

	void SavePosition();
	void LoadPosition();

	void OnTextEditorUpdated();

	static INT_PTR CALLBACK StaticAboutDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);

	enum {
		kIdTimerCompile = 100
	};

	HMENU mhmenu;
	HFONT mhfont;
	HACCEL	mhAccelTable;
	HWND	mhwndShaderEditor;
	vdrefptr<IVDTextEditor> mpShaderEditor;
	HWND	mhwndLog;
	vdrefptr<IVDTextEditor> mpLogEditor;
	UINT	mReloadTimerId;
	bool	mbExit;
	bool	mbModified;
	bool	mbBlockChanges;

	int		mErrorLineIndex;

	VDShaderEditorSplitterBar	*mpSplitterBar;

	VDShaderFilterConfig& mConfig;
	IVDFilterPreview *mifp;
	IVDFilterPreview2 *mifp2;

	struct PresetEntry {
		const char *mpText;
		size_t		mLength;
	};

	vdfastvector<PresetEntry> mPresets;

	typedef std::list<VDStringA> MRUList;
	MRUList mMRUList;

	LOGFONT	mFontDesc;

	UINT	mFindMsgId;
	HWND	mhwndFind;
	FINDREPLACE	mFindReplace;

	char	mFindBuffer[2048];
	char	mReplaceBuffer[2048];
};

///////////////////////////////////////////////////////////////////////////////
VDShaderEditor::VDShaderEditor(VDShaderFilterConfig& config, IVDFilterPreview *ifp, IVDFilterPreview2 *ifp2)
	: mhmenu(NULL)
	, mhfont(NULL)
	, mhAccelTable(NULL)
	, mhwndShaderEditor(NULL)
	, mhwndLog(NULL)
	, mReloadTimerId(0)
	, mbExit(false)
	, mbModified(false)
	, mbBlockChanges(false)
	, mErrorLineIndex(-1)
	, mConfig(config)
	, mifp(ifp)
	, mifp2(ifp2)
	, mFindMsgId(0)
	, mhwndFind(NULL)
{
	mpSplitterBar = new VDShaderEditorSplitterBar;
	mpSplitterBar->AddRef();

	VDRegistryAppKey key;
	mFontDesc.lfHeight			= key.getInt("Font size", -14);
	mFontDesc.lfWidth			= 0;
	mFontDesc.lfEscapement		= 0;
	mFontDesc.lfOrientation		= 0;
	mFontDesc.lfWidth			= 0;
	mFontDesc.lfItalic			= FALSE;
	mFontDesc.lfUnderline		= FALSE;
	mFontDesc.lfStrikeOut		= FALSE;
	mFontDesc.lfCharSet			= ANSI_CHARSET;
	mFontDesc.lfOutPrecision	= OUT_DEFAULT_PRECIS;
	mFontDesc.lfClipPrecision	= CLIP_DEFAULT_PRECIS;
	mFontDesc.lfQuality			= DEFAULT_QUALITY;
	mFontDesc.lfPitchAndFamily	= DEFAULT_PITCH | FF_DONTCARE;

	VDStringA fontname;
	if (key.getString("Font name", fontname))
		vdstrlcpy(mFontDesc.lfFaceName, fontname.c_str(), sizeof(mFontDesc.lfFaceName)/sizeof(mFontDesc.lfFaceName[0]));
	else
		vdstrlcpy(mFontDesc.lfFaceName, "Lucida Console", sizeof(mFontDesc.lfFaceName)/sizeof(mFontDesc.lfFaceName[0]));

	mFindBuffer[0] = 0;
	mReplaceBuffer[0] = 0;
	mFindReplace.lStructSize		= sizeof(mFindReplace);
	mFindReplace.hwndOwner			= NULL;
	mFindReplace.hInstance			= VDGetLocalModuleHandleW32();
	mFindReplace.Flags				= FR_DOWN;
	mFindReplace.lpstrFindWhat		= mFindBuffer;
	mFindReplace.lpstrReplaceWith	= mReplaceBuffer;
	mFindReplace.wFindWhatLen		= sizeof(mFindBuffer);
	mFindReplace.wReplaceWithLen	= sizeof(mReplaceBuffer);
	mFindReplace.lCustData			= (LPARAM)this;
	mFindReplace.lpfnHook			= NULL;
	mFindReplace.lpTemplateName		= NULL;
}

VDShaderEditor::~VDShaderEditor() {
	mpSplitterBar->Release();
}

bool VDShaderEditor::Display(HWND hwndParent) {
	HINSTANCE hinstRE20 = LoadLibrary("riched20");
	if (hinstRE20) {
		mConfig.mpSource = this;

		bool parentWasEnabled = false;
		if (hwndParent)
			parentWasEnabled = !EnableWindow(hwndParent, FALSE);

		HWND hwnd = CreateWindow((LPCTSTR)sWndClass, "VirtualDub shader editor", WS_OVERLAPPEDWINDOW|WS_VISIBLE, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, hwndParent, NULL, (HINSTANCE)&__ImageBase, static_cast<VDShaderEditorBaseWindow *>(this));

		if (hwnd) {
			MSG msg;

			msg.message = 0;

			while(!mbExit && GetMessage(&msg, NULL, 0, 0)) {
				if (CallMsgFilter(&msg, 0))
					continue;

				if (mhwndFind && IsDialogMessage(mhwndFind, &msg))
					continue;

				if (mhAccelTable && TranslateAccelerator(hwnd, mhAccelTable, &msg))
					continue;

				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			if (parentWasEnabled)
				EnableWindow(hwndParent, TRUE);

			if (mhwnd) {
				VDASSERT(IsWindow(mhwnd));
				VDASSERT(!(GetWindowLong(hwndParent, GWL_STYLE) & WS_DISABLED));
				DestroyWindow(mhwnd);
				VDASSERT(GetForegroundWindow() == hwndParent);
			}

			if (msg.message == WM_QUIT)
				PostQuitMessage((int)msg.wParam);
		}

		mConfig.mpSource = NULL;
	}
	FreeLibrary(hinstRE20);

	return true;
}

namespace {
	class StringStreamOut : public IVDTextEditorStreamOut {
	public:
		StringStreamOut(VDStringA& dst) : mDst(dst) {}

		void Write(const void *buf, size_t len) {
			mDst.append((const char *)buf, (const char *)buf+len);
		}

	protected:
		VDStringA& mDst;
	};
}

void VDShaderEditor::ReadSource(VDStringA& dst) {
	dst.clear();

	mpShaderEditor->Save(StringStreamOut(dst));
}

LRESULT VDShaderEditor::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_CREATE:
			OnCreate();
			break;

		case WM_DESTROY:
			OnDestroy();
			return 0;

		case WM_NCDESTROY:
			OnNCDestroy();
			break;

		case WM_CLOSE:
			DoFileExit();
			return 0;

		case WM_SIZE:
			OnSize();
			return 0;

		case WM_COMMAND:
			OnCommand(LOWORD(wParam));
			return 0;

		case WM_TIMER:
			if (OnTimer(LOWORD(wParam)))
				return 0;
			break;

		case WM_INITMENU:
			if ((HMENU)wParam == mhmenu) {
				OnInitMenu();
				return 0;
			}
			break;

		case WM_SETFOCUS:
			if (!mbExit) {
				SetFocus(mhwndShaderEditor);
				return 0;
			}
			break;

		default:
			if (msg == mFindMsgId && msg) {
				OnFindMsg();
				return 0;
			}
			break;
	}

	return VDShaderEditorBaseWindow::WndProc(msg, wParam, lParam);
}

void VDShaderEditor::OnCreate() {
	mFindMsgId = RegisterWindowMessage(FINDMSGSTRING);
	mhAccelTable = LoadAccelerators((HINSTANCE)&__ImageBase, MAKEINTRESOURCE(IDR_ACCEL_EDITOR));

	mhmenu = LoadMenu((HINSTANCE)&__ImageBase, MAKEINTRESOURCE(IDR_MENU_EDITOR));
	SetMenu(mhwnd, mhmenu);

	LoadMRUList();
	UpdateMRUListMenu();

	LoadPresets();
	LoadPosition();

	CreateWindow((LPCTSTR)sWndClass, "", WS_CHILD|WS_VISIBLE, 0, 0, 0, 0, mhwnd, NULL, (HINSTANCE)&__ImageBase, static_cast<VDShaderEditorBaseWindow *>(mpSplitterBar));

	VDCreateTextEditor(~mpShaderEditor);
	mpShaderEditor->SetCallback(this);
	mhwndShaderEditor = (HWND)mpShaderEditor->Create(0, WS_CHILD|WS_VSCROLL|WS_VISIBLE, 0, 0, 0, 0, (VDGUIHandle)mpSplitterBar->GetHandleW32(), 100);

	VDCreateTextEditor(~mpLogEditor);
	mpLogEditor->SetReadOnly(true);
	mhwndLog = (HWND)mpLogEditor->Create(0, WS_CHILD|WS_VSCROLL|WS_VISIBLE, 0, 0, 0, 0, (VDGUIHandle)mpSplitterBar->GetHandleW32(), 101);

	mhfont = CreateFontIndirect(&mFontDesc);
	if (mhfont) {
		SendMessage(mhwndShaderEditor, WM_SETFONT, (WPARAM)mhfont, TRUE);
		SendMessage(mhwndLog, WM_SETFONT, (WPARAM)mhfont, TRUE);
	}

	OnSize();
	UpdateCaption();

	if (mConfig.mFilename[0]) {
		if (GetFileAttributes(mConfig.mFilename) != 0xFFFFFFFF)
			DoFileOpen(mConfig.mFilename);
	}
}

void VDShaderEditor::OnDestroy() {
	SaveMRUList();
	SavePosition();
}

void VDShaderEditor::OnNCDestroy() {
	if (mhfont) {
		DeleteObject(mhfont);
		mhfont = NULL;
	}

	if (mhmenu) {
		SetMenu(mhwnd, NULL);
		DestroyMenu(mhmenu);
		mhmenu = NULL;
	}
}

void VDShaderEditor::OnSize() {
	RECT r;
	GetClientRect(mhwnd, &r);

	SetWindowPos(mpSplitterBar->GetHandleW32(), NULL, 0, 0, r.right, r.bottom, SWP_NOMOVE|SWP_NOZORDER|SWP_NOACTIVATE);
}

bool VDShaderEditor::OnTimer(int id) {
	if (id == kIdTimerCompile) {
		KillTimer(mhwnd, mReloadTimerId);
		mReloadTimerId = 0;
		Compile();
		return true;
	}

	return false;
}

void VDShaderEditor::OnInitMenu() {
	IVDTextEditor *editor = mpShaderEditor;
	if (GetFocus() == mhwndLog)
		editor = mpLogEditor;

	VDEnableMenuItemByCommandW32(mhmenu, ID_EDIT_CUT, editor->IsCutPossible());
	VDEnableMenuItemByCommandW32(mhmenu, ID_EDIT_COPY, editor->IsCopyPossible());

	VDEnableMenuItemByCommandW32(mhmenu, ID_EDIT_PASTE, editor->IsPastePossible());
	VDEnableMenuItemByCommandW32(mhmenu, ID_EDIT_UNDO, editor->IsUndoPossible());
	VDEnableMenuItemByCommandW32(mhmenu, ID_EDIT_REDO, editor->IsRedoPossible());

	VDEnableMenuItemByCommandW32(mhmenu, ID_OPTIONS_PREVIEWWINDOW, mifp && mifp->isPreviewEnabled());
	VDCheckMenuItemByCommandW32(mhmenu, ID_OPTIONS_PREVIEWWINDOW, mifp2 && mifp2->IsPreviewDisplayed());
	VDCheckMenuItemByCommandW32(mhmenu, ID_OPTIONS_OPTIMIZESHADER, mConfig.mbOptimizationEnabled);

	VDCheckMenuItemByCommandW32(mhmenu, ID_EXECUTIONMODE_DEFAULT, mConfig.mExecutionMode == VDShaderFilterConfig::kModeDefault);
	VDCheckMenuItemByCommandW32(mhmenu, ID_EXECUTIONMODE_INTERPRETER, mConfig.mExecutionMode == VDShaderFilterConfig::kModeInterp);
	VDCheckMenuItemByCommandW32(mhmenu, ID_EXECUTIONMODE_SCALARJIT, mConfig.mExecutionMode == VDShaderFilterConfig::kModeScalarJIT);
}

bool VDShaderEditor::OnCommand(int cmd) {
	switch(cmd) {
		case ID_FILE_NEW:
			DoFileNew();
			return true;

		case ID_FILE_OPEN:
			DoFileOpen();
			return true;

		case ID_FILE_SAVE:
			DoFileSave();
			return true;

		case ID_FILE_SAVEAS:
			DoFileSaveAs();
			return true;

		case ID_FILE_EXIT:
			DoFileExit();
			return true;

		case ID_VIEW_TEXTEDITOR:
			SetFocus(mhwndShaderEditor);
			return true;

		case ID_EDIT_UNDO:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Undo();
			else
				mpShaderEditor->Undo();
			return true;

		case ID_EDIT_REDO:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Redo();
			else
				mpShaderEditor->Redo();
			return true;

		case ID_EDIT_CUT:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Cut();
			else
				mpShaderEditor->Cut();
			return true;

		case ID_EDIT_COPY:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Copy();
			else
				mpShaderEditor->Copy();
			return true;

		case ID_EDIT_PASTE:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Paste();
			else
				mpShaderEditor->Paste();
			return true;

		case ID_EDIT_DELETE:
			if (GetFocus() == mhwndLog)
				mpLogEditor->Delete();
			else
				mpShaderEditor->Delete();
			return true;

		case ID_EDIT_SELECTALL:
			if (GetFocus() == mhwndLog)
				mpLogEditor->SelectAll();
			else
				mpShaderEditor->SelectAll();
			return true;

		case ID_EDIT_FIND:
			DoEditFind();
			return true;

		case ID_EDIT_FINDNEXT:
			DoEditFindNext();
			return true;

		case ID_EDIT_NEXTERROR:
			DoEditNextError();
			return true;

		case ID_EDIT_PREVERROR:
			DoEditPrevError();
			return true;

		case ID_BUILD_COMPILE:
			Compile();
			return true;

		case ID_OPTIONS_PREVIEWWINDOW:
			if (mifp)
				mifp->Toggle(mhwnd);
			return true;

		case ID_OPTIONS_OPTIMIZESHADER:
			mConfig.mbOptimizationEnabled = !mConfig.mbOptimizationEnabled;
			Compile();
			if (mifp)
				mifp->RedoSystem();
			return true;

		case ID_OPTIONS_FONT:
			DoOptionsChooseFont();
			return true;

		case ID_EXECUTIONMODE_DEFAULT:
			mConfig.mExecutionMode = VDShaderFilterConfig::kModeDefault;
			if (mifp)
				mifp->RedoSystem();
			return true;
		case ID_EXECUTIONMODE_INTERPRETER:
			mConfig.mExecutionMode = VDShaderFilterConfig::kModeInterp;
			if (mifp)
				mifp->RedoSystem();
			return true;
		case ID_EXECUTIONMODE_SCALARJIT:
			mConfig.mExecutionMode = VDShaderFilterConfig::kModeScalarJIT;
			if (mifp)
				mifp->RedoSystem();
			return true;

		case ID_HELP_TUTORIAL:
			OpenPreset(0);
			return true;

		case ID_HELP_ABOUT:
			DialogBox(VDGetLocalModuleHandleW32(), MAKEINTRESOURCE(IDD_ABOUT), mhwnd, StaticAboutDlgProc);
			return true;

		default:
			{
				uint32 index = (uint32)(cmd - ID_PRESET_BASE);
				if (index < 100) {
					OpenPreset(index);
					return true;
				}

				index = (uint32)(cmd - ID_MRU_BASE);
				if (index < mMRUList.size()) {
					MRUList::iterator it(mMRUList.begin());
					std::advance(it, index);
					if (QuerySaveFirst())
						DoFileOpen(it->c_str());
					return true;
				}
			}
			break;
	}

	return false;
}

void VDShaderEditor::OnFindMsg() {
	if (mFindReplace.Flags & FR_DIALOGTERM)
		mhwndFind = NULL;
	else if (mFindReplace.Flags & FR_FINDNEXT)
		DoEditFindNext();
}

void VDShaderEditor::WriteLogOutput(const char *s) {
	mpLogEditor->Append(s);
}

void VDShaderEditor::Compile() {
	std::vector<char> text;
	int len = GetWindowTextLength(mhwndShaderEditor);
	text.resize(len + 1);
	if (len)
		len = GetWindowText(mhwndShaderEditor, &text[0], len + 1);

	mErrorLineIndex = -1;

	VDShaderParser parser;

	SetWindowText(mhwndLog, "");
	vdrefptr<VDShaderEffect> effect;
	if (parser.Parse(&text[0], len, mConfig.mFilename[0] ? mConfig.mFilename : "<unnamed file>", mConfig.mbOptimizationEnabled, this, ~effect)) {
		VDShaderDisassembleEffect(*this, *effect);

		const VDShaderTechnique *tech = effect->GetTechniqueByIndex(0);
		if (tech) {
			const VDShaderPass *pass = tech->GetPassByIndex(0);
			if (pass) {
				const VDShaderNameTable& nameTable = effect->NameTable();
				VDShaderBindingElement binding[1]={
					{ 0, nsVDShaderParser::kTypeFloat4, nameTable.GetName("COLOR0", 6) }
				};

				WriteLogOutput("\n");

				if (mConfig.mExecutionMode == VDShaderFilterConfig::kModeInterp) {
					VDShaderInterpreter interp;
					interp.LoadShader(*effect->GetFunctionByIndex(pass->GetPixelShaderIndex()), binding, 1, this);
				} else {
					VDShaderScalarJIT jit;
					VDShaderInterpreterSampler samp;
					samp.Init(NULL, 0, 1, 1, true, true, false);
					jit.LoadSamplers(&samp, 1);
					jit.LoadShader(*effect->GetFunctionByIndex(pass->GetPixelShaderIndex()), binding, 1, this);
				}
			}
		}

		mConfig.mpEffect = effect;
		if (mifp)
			mifp->RedoSystem();

		effect = NULL;
	}
}

void VDShaderEditor::UpdateCaption() {
	char buf[512];

	if (mConfig.mFilename[0]) {
		if ((unsigned)_snprintf(buf, 512, "VirtualDub shader editor - [%s]%s", mConfig.mFilename, mbModified ? "*" : "") < 512)
			SetWindowText(mhwnd, buf);
	} else {
		if ((unsigned)_snprintf(buf, 512, "VirtualDub shader editor - New file%s", mbModified ? "*" : "") < 512)
			SetWindowText(mhwnd, buf);
	}
}

void VDShaderEditor::DoFileNew() {
	if (!QuerySaveFirst())
		return;

	SetWindowText(mhwndShaderEditor, "");
	mbModified = false;
	mConfig.mFilename[0] = 0;
	UpdateCaption();
}

namespace {
	DWORD CALLBACK StreamIn(DWORD_PTR dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb) {
		HANDLE h = (HANDLE)dwCookie;

		DWORD actual;
		if (ReadFile(h, pbBuff, cb, &actual, NULL)) {
			if (pcb)
				*pcb = (LONG)actual;

			return 0;
		} else
			return GetLastError();
	}
}

void VDShaderEditor::DoFileOpen() {
	if (!QuerySaveFirst())
		return;

	OPENFILENAME ofn = {};

#ifdef OPENFILENAME_SIZE_VERSION_400
	ofn.lStructSize		= OPENFILENAME_SIZE_VERSION_400;
#else
	ofn.lStructSize		= sizeof(OPENFILENAME);
#endif

	ofn.hwndOwner		= mhwnd;
	ofn.hInstance		= (HINSTANCE)&__ImageBase;
	ofn.lpstrFilter		= "Effect file (*.fx)\0*.fx\0All files\0*.*\0";

	char fnbuf[MAX_PATH];
	vdstrlcpy(fnbuf, mConfig.mFilename, MAX_PATH);
	ofn.lpstrFile		= fnbuf;
	ofn.nMaxFile		= MAX_PATH;
	ofn.lpstrTitle		= "Open shader file";
	ofn.Flags			= OFN_HIDEREADONLY | OFN_FILEMUSTEXIST;
	ofn.lpstrDefExt		= "fx";

	if (GetOpenFileName(&ofn)) {
		DoFileOpen(fnbuf);
	}
}

namespace {
	class CaseInsensitiveEquals {
	public:
		CaseInsensitiveEquals(const char *s) : mpString(s) {}

		bool operator()(const VDStringA& s) const {
			return !_stricmp(s.c_str(), mpString);
		}

	private:
		const char *mpString;
	};
}

void VDShaderEditor::DoFileOpen(const char *s) {
	try {
		VDFileStream fs(s);

		mpShaderEditor->Load(fs);

		VDString tmp(s);		// in case it aliases mFilename
		vdstrlcpy(mConfig.mFilename, tmp.c_str(), sizeof(mConfig.mFilename)/sizeof(mConfig.mFilename[0]));

		MRUList::iterator it(std::find_if(mMRUList.begin(), mMRUList.end(), CaseInsensitiveEquals(mConfig.mFilename)));

		if (it != mMRUList.end()) {
			if (it != mMRUList.begin())
				mMRUList.splice(mMRUList.begin(), mMRUList, it);
		} else {
			if (mMRUList.size() < 4)
				mMRUList.push_front(tmp);
			else {
				--it;
				mMRUList.splice(mMRUList.begin(), mMRUList, it);
				*it = tmp;
			}
		}

		UpdateMRUListMenu();
		mbModified = false;
	} catch(const MyError& e) {
		MessageBox(mhwnd, e.gets(), "Error reading file", MB_ICONERROR|MB_OK);
	}

	SetFocus(mhwndShaderEditor);
	UpdateCaption();
}

namespace {
	class TextEditorWriterAdapter : public IVDTextEditorStreamOut {
	public:
		TextEditorWriterAdapter(VDFile& f) : mFile(f) {}

		void Write(const void *buf, size_t len);

	protected:
		VDFile& mFile;
	};

	void TextEditorWriterAdapter::Write(const void *buf, size_t len) {
		mFile.write(buf, len);
	}
}

void VDShaderEditor::DoFileSave() {
	if (mConfig.mFilename[0] == 0)
		return DoFileSaveAs();

	try {
		VDFile fs(mConfig.mFilename, nsVDFile::kWrite | nsVDFile::kDenyRead | nsVDFile::kCreateAlways);
		TextEditorWriterAdapter adapter(fs);
		mpShaderEditor->Save(adapter);
		mbModified = false;
		UpdateCaption();
	} catch(const MyError& e) {
		MessageBox(mhwnd, e.gets(), "Error saving file", MB_OK|MB_ICONERROR);
	}
}

void VDShaderEditor::DoFileSaveAs() {
	OPENFILENAME ofn = {};

#ifdef OPENFILENAME_SIZE_VERSION_400
	ofn.lStructSize		= OPENFILENAME_SIZE_VERSION_400;
#else
	ofn.lStructSize		= sizeof(OPENFILENAME);
#endif

	ofn.hwndOwner		= mhwnd;
	ofn.hInstance		= (HINSTANCE)&__ImageBase;
	ofn.lpstrFilter		= "Effect file (*.fx)\0*.fx\0All files\0*.*\0";
	ofn.lpstrFile		= mConfig.mFilename;
	ofn.nMaxFile		= MAX_PATH;
	ofn.lpstrTitle		= "Save shader file";
	ofn.Flags			= OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt		= "fx";

	if (GetSaveFileName(&ofn)) {
		DoFileSave();
	}
	SetFocus(mhwndShaderEditor);
}

void VDShaderEditor::DoFileExit() {
	if (!QuerySaveFirst())
		return;

	if (mhwndFind) {
		DestroyWindow(mhwndFind);
		mhwndFind = NULL;
	}

	if (!mbExit) {
		mbExit = true;

		// For some reason this is necessary to avoid losing focus to another app right after
		// displaying a message box... dunno why.
		HWND hwndParent = GetWindow(mhwnd, GW_OWNER);
		if (hwndParent)
			SetForegroundWindow(hwndParent);
		PostMessage(mhwnd, WM_NULL, 0, 0);
	}
}

void VDShaderEditor::DoEditFind() {
	if (mhwndFind)
		SetForegroundWindow(mhwndFind);
	else {
		mFindReplace.hwndOwner = mhwnd;
		mFindReplace.Flags &= ~(FR_FINDNEXT | FR_REPLACE | FR_REPLACEALL | FR_DIALOGTERM);
		mhwndFind = FindText(&mFindReplace);
	}
}

void VDShaderEditor::DoEditFindNext() {
	bool matchCase = 0 != (mFindReplace.Flags & FR_MATCHCASE);
	bool wholeWord = 0 != (mFindReplace.Flags & FR_WHOLEWORD);
	bool searchUp = !(mFindReplace.Flags & FR_DOWN);

	if (!mpShaderEditor->Find(mFindReplace.lpstrFindWhat, strlen(mFindReplace.lpstrFindWhat), matchCase, wholeWord, searchUp)) {
		HWND hwndFore = GetForegroundWindow();

		MessageBox(hwndFore == mhwndFind ? mhwndFind : mhwnd, "Search string not found.", "VDShader Editor", MB_OK | MB_ICONINFORMATION);
	}
}

void VDShaderEditor::DoEditPrevError() {
	int lines = mpLogEditor->GetLineCount();

	for(int i = mErrorLineIndex+1; i<lines; ++i) {
		if (HighlightError(i)) {
			mErrorLineIndex = i;
			return;
		}
	}

	for(int i = 0; i <= mErrorLineIndex; ++i) {
		if (HighlightError(i)) {
			mErrorLineIndex = i;
			return;
		}
	}
}

void VDShaderEditor::DoEditNextError() {
	int lines = mpLogEditor->GetLineCount();

	for(int i = mErrorLineIndex-1; i>=0; --i) {
		if (HighlightError(i)) {
			mErrorLineIndex = i;
			return;
		}
	}

	for(int i = lines-1; i >= mErrorLineIndex; --i) {
		if (HighlightError(i)) {
			mErrorLineIndex = i;
			return;
		}
	}
}

void VDShaderEditor::DoOptionsChooseFont() {
	CHOOSEFONT cf={};

	cf.lStructSize		= sizeof(CHOOSEFONT);
	cf.hwndOwner		= mhwnd;
	cf.hDC				= NULL;
	cf.lpLogFont		= &mFontDesc;
	cf.iPointSize		= 0;
	cf.Flags			= CF_FORCEFONTEXIST | CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
	cf.rgbColors		= 0;
	cf.lCustData		= 0;
	cf.lpfnHook			= NULL;
	cf.lpTemplateName	= NULL;
	cf.hInstance		= VDGetLocalModuleHandleW32();
	cf.lpszStyle		= NULL;
	cf.nFontType		= REGULAR_FONTTYPE;
	cf.nSizeMin			= 0;
	cf.nSizeMax			= 0;

	if (ChooseFont(&cf)) {
		HFONT hf = CreateFontIndirect(&mFontDesc);
		if (hf) {
			SendMessage(mhwndShaderEditor, WM_SETFONT, (WPARAM)hf, TRUE);
			SendMessage(mhwndLog, WM_SETFONT, (WPARAM)hf, TRUE);
			if (mhfont)
				DeleteObject(mhfont);
			mhfont = hf;

			VDRegistryAppKey key;
			key.setInt("Font size", mFontDesc.lfHeight);
			key.setString("Font name", mFontDesc.lfFaceName);
		}
	}
}

bool VDShaderEditor::HighlightError(int line) {
	vdfastvector<char> buf;

	if (mpLogEditor->GetLineText(line, buf)) {
		buf.push_back(0);

		int errline, errcol;
		if (2 != sscanf(buf.data(), "%*[^(](%d,%d):", &errline, &errcol))
			return false;

		if (errline < 1)
			errline = 1;

		mpShaderEditor->SetCursorPos(errline - 1, errcol - 1);
	}

	return true;
}

bool VDShaderEditor::QuerySaveFirst() {
	if (mbModified) {
		DWORD id = MessageBox(mhwnd, "The current document has been modified and will be lost. Save first?", "Shader editor warning", MB_ICONEXCLAMATION | MB_YESNOCANCEL);

		if (id == IDCANCEL)
			return false;

		if (id == IDYES)
			DoFileSave();
	}

	return true;
}

namespace {
	DWORD CALLBACK StringStreamIn(DWORD_PTR dwCookie, LPBYTE pbBuff, LONG cb, LONG *pcb) {
		VDStringSpanA *pSpan = (VDStringSpanA *)dwCookie;

		LONG actual = (LONG)pSpan->copy((char *)pbBuff, (VDStringSpanA::size_type)cb);
		if (pcb)
			*pcb = actual;

		if (actual)
			*pSpan = pSpan->subspan(actual);

		return 0;
	}
}

void VDShaderEditor::OpenPreset(int index) {
	if ((uint32)index >= mPresets.size())
		return;

	const PresetEntry& preset = mPresets[index];

	if (!QuerySaveFirst())
		return;

	mConfig.mFilename[0] = 0;

	try {
		VDMemoryStream ms(preset.mpText, preset.mLength);
		mpShaderEditor->Load(ms);
		mbModified = false;
		UpdateCaption();
	} catch(const MyError& e) {
		MessageBox(mhwnd, e.gets(), "Error loading preset", MB_OK | MB_ICONERROR);
	}
}

void VDShaderEditor::LoadPresets() {
	HMENU hPresetMenu = CreatePopupMenu();

	MENUITEMINFO mii;
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400A;
	mii.fMask = MIIM_SUBMENU;
	mii.hSubMenu = hPresetMenu;
	SetMenuItemInfo(mhmenu, 4, TRUE, &mii);

	HINSTANCE hInst = VDGetLocalModuleHandleW32();
	LoadPreset(hInst, "TUTORIAL", MAKEINTRESOURCE(IDR_TUTORIAL), false);
	EnumResourceNames(hInst, "SAMPLE", StaticLoadPreset, (LONG_PTR)this);
}

BOOL CALLBACK VDShaderEditor::StaticLoadPreset(HMODULE hModule, LPCTSTR lpszType, LPTSTR lpszName, LONG_PTR lParam) {
	((VDShaderEditor *)lParam)->LoadPreset(hModule, lpszType, lpszName, true);
	return TRUE;
}

void VDShaderEditor::LoadPreset(HMODULE hModule, LPCTSTR lpszType, LPTSTR lpszName, bool addToMenu) {
	HRSRC hResInfo = FindResource(hModule, lpszName, lpszType);
	if (hResInfo) {
		HGLOBAL hResData = LoadResource(hModule, hResInfo);
		if (hResData) {
			void *data = LockResource(hResData);
			if (data) {
				DWORD len = SizeofResource(hModule, hResInfo);
				if (len) {
					PresetEntry ent;
					ent.mpText = (const char *)data;
					ent.mLength = len;

					mPresets.push_back(ent);

					if (mhmenu && addToMenu) {
						HMENU hMenuPresets = GetSubMenu(mhmenu, 4);

						VDStringSpanA ssa(ent.mpText, ent.mpText + ent.mLength);
						VDStringSpanA::size_type endpos = ssa.find('\n');

						if (endpos >= 3) {
							VDStringA name(ssa.data() + 3, endpos - 3);
							AppendMenu(hMenuPresets, MF_STRING|MF_ENABLED, ID_PRESET_BASE + (mPresets.size() - 1), name.c_str());
						}
					}
				}
			}
		}
	}
}

void VDShaderEditor::LoadMRUList() {
	mMRUList.clear();
	VDRegistryAppKey key("MRU list");
	char buf[]="a";

	VDStringA fn;
	while(key.getString(buf, fn)) {
		mMRUList.push_back(fn);
		++buf[0];
	}
}

void VDShaderEditor::SaveMRUList() {
	VDRegistryAppKey key("MRU list");
	char buf[]="a";

	MRUList::const_iterator it(mMRUList.begin()), itEnd(mMRUList.end());
	for(; it!=itEnd; ++it) {
		key.setString(buf, it->c_str());
		++buf[0];
	}

	key.removeValue(buf);
}

void VDShaderEditor::UpdateMRUListMenu() {
	HMENU hmenuFile = GetSubMenu(mhmenu, 0);
	if (!hmenuFile)
		return;

	int idx = GetMenuItemCount(hmenuFile) - 3;

	MENUITEMINFO mii;
	while(idx >= 0) {
		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400A;
		mii.fMask = MIIM_FTYPE;
		if (!GetMenuItemInfo(hmenuFile, idx, TRUE, &mii))
			return;

		if (mii.fType & MFT_SEPARATOR) {
			++idx;
			break;
		}

		RemoveMenu(hmenuFile, idx, MF_BYPOSITION);
		--idx;
	}

	MRUList::const_iterator it(mMRUList.begin()), itEnd(mMRUList.end());
	VDStringA buf;
	int id = ID_MRU_BASE;
	for(; it!=itEnd; ++it) {
		const VDStringA& s = *it;

		buf.sprintf("&%d %s", id + 1 - ID_MRU_BASE, s.c_str());

		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400A;
		mii.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_STATE | MIIM_ID;
		mii.fType = MFT_STRING;
		mii.fState = MFS_ENABLED;
		mii.wID = id++;
		mii.dwTypeData = (LPSTR)buf.c_str();
		mii.cch = buf.size();
		InsertMenuItem(hmenuFile, idx++, TRUE, &mii);
	}
}

void VDShaderEditor::SavePosition() {
	WINDOWPLACEMENT wp = {sizeof(WINDOWPLACEMENT)};
	if (GetWindowPlacement(mhwnd, &wp)) {
		VDRegistryAppKey key("Window position");
		key.setInt("X", wp.rcNormalPosition.left);
		key.setInt("Y", wp.rcNormalPosition.top);
		key.setInt("Width", wp.rcNormalPosition.right - wp.rcNormalPosition.left);
		key.setInt("Height", wp.rcNormalPosition.bottom - wp.rcNormalPosition.top);
		key.setBool("Maximized", wp.showCmd == SW_SHOWMAXIMIZED);
	}
}

void VDShaderEditor::LoadPosition() {
	WINDOWPLACEMENT wp = {sizeof(WINDOWPLACEMENT)};
	if (GetWindowPlacement(mhwnd, &wp)) {
		VDRegistryAppKey key("Window position");

		int x = key.getInt("X", INT_MIN);
		int y = key.getInt("Y", INT_MIN);
		int w = key.getInt("Width", INT_MIN);
		int h = key.getInt("Height", INT_MIN);

		if (w == INT_MIN || h == INT_MIN) {
			w = wp.rcNormalPosition.right - wp.rcNormalPosition.left;
			h = wp.rcNormalPosition.bottom - wp.rcNormalPosition.top;
		}

		if (x == INT_MIN || y == INT_MIN) {
			x = wp.rcNormalPosition.left;
			y = wp.rcNormalPosition.top;
		}

		wp.rcNormalPosition.left = x;
		wp.rcNormalPosition.top = y;
		wp.rcNormalPosition.right = x + w;
		wp.rcNormalPosition.bottom = y + h;

		if (key.getBool("Maximized", false))
			wp.showCmd = SW_MAXIMIZE;

		SetWindowPlacement(mhwnd, &wp);
	}
}

void VDShaderEditor::OnTextEditorUpdated() {
	mReloadTimerId = SetTimer(mhwnd, kIdTimerCompile, 700, NULL);
	if (!mbModified) {
		mbModified = true;
		UpdateCaption();
	}
}

INT_PTR CALLBACK VDShaderEditor::StaticAboutDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDOK:
			case IDCANCEL:
				EndDialog(hdlg, 0);
				return TRUE;
			}
			break;
	}
	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
bool VDOpenShaderEditor(HWND hwnd, VDShaderFilterConfig& config, IVDFilterPreview *ifp, IVDFilterPreview2 *ifp2) {
	VDShaderEditor *ed = new VDShaderEditor(config, ifp, ifp2);
	ed->AddRef();
	ed->Display(hwnd);
	ed->Release();
	return true;
}

extern "C" __declspec(dllexport) void CALLBACK OpenEditor(HWND hwnd, HINSTANCE hinst, LPSTR lpszCmdLine, int nCmdShow) {
	VDRegistryAppKey::setDefaultKey("Software\\Freeware\\VDShader\\");
	VDRegisterShaderEditor();
	VDShaderFilterConfig config;
	VDOpenShaderEditor(hwnd, config, NULL, NULL);
	VDUnregisterShaderEditor();
}

void VDRegisterShaderEditor() {
	VDShaderEditorBaseWindow::Register();
}

void VDUnregisterShaderEditor() {
	VDShaderEditorBaseWindow::Unregister();
}
