//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <vector>
#include <float.h>
#include "parser.h"
#include "interp.h"
#include "log.h"
#include "disasm.h"

///////////////////////////////////////////////////////////////////////////////
void VDShaderInterpreterSampler::Init(const uint32 *src, ptrdiff_t pitch, uint32 w, uint32 h, bool clampU, bool clampV, bool bilinear) {
	mpSrc		= src;
	mPitch		= pitch;
	mFWidth		= (float)w;
	mFHeight	= (float)h;
	mULimit		= w - 1;
	mVLimit		= h - 1;
	mbClampU	= clampU;
	mbClampV	= clampV;
	mbBilinear	= bilinear;

	mFWidth0 = mFWidth;
	mFHeight0 = mFHeight;
	if (bilinear) {
		mFWidth *= 64.0f;
		mFHeight *= 64.0f;
		mULimit <<= 6;
		mVLimit <<= 6;
	}
}

uint32 VDShaderInterpreterSampler::Sample(float u, float v) const {
	if (!mbClampU)
		u -= floorf(u);

	if (!mbClampV)
		v -= floorf(v);

	u *= mFWidth;
	v *= mFHeight;

	if (mbBilinear) {
		int iu = (int)u;
		int iv = (int)v;

		if (mbClampU && (unsigned)iu > mULimit)
			iu = ~(iu >> 31) & mULimit;
		if (mbClampV && (unsigned)iv > mVLimit)
			iv = ~(iv >> 31) & mVLimit;

		iu -= 32;
		iv -= 32;

		const uint32 *row0 = (const uint32 *)((const char *)mpSrc + mPitch*(iv >> 6)) + (iu >> 6);
		const uint32 *row1 = (const uint32 *)((const char *)row0 + mPitch);
		uint32 px0 = row0[0];
		uint32 px1 = row0[1];
		uint32 px2 = row1[0];
		uint32 px3 = row1[1];

		uint32 px0_rb = px0 & 0xff00ff;
		uint32 px1_rb = px1 & 0xff00ff;
		uint32 px2_rb = px2 & 0xff00ff;
		uint32 px3_rb = px3 & 0xff00ff;
		uint32 px0_ag = px0 & 0xff00ff00;
		uint32 px1_ag = px1 & 0xff00ff00;
		uint32 px2_ag = px2 & 0xff00ff00;
		uint32 px3_ag = px3 & 0xff00ff00;

		sint32 uf = iu & 63;
		sint32 vf = iv & 63;

		uint32 px01_rb = (px0_rb + (((sint32)(px1_rb - px0_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px23_rb = (px2_rb + (((sint32)(px3_rb - px2_rb)*uf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px01_ag = (px0_ag + ((sint32)((px1_ag >> 6) - (px0_ag >> 6))*uf + 0x800080)) & 0xff00ff00;
		uint32 px23_ag = (px2_ag + ((sint32)((px3_ag >> 6) - (px2_ag >> 6))*uf + 0x800080)) & 0xff00ff00;

		uint32 px_rb = (px01_rb + (((sint32)(px23_rb - px01_rb)*vf + 0x200020) >> 6)) & 0xff00ff;
		uint32 px_ag = (px01_ag + ((sint32)((px23_ag >> 6) - (px01_ag >> 6))*vf + 0x800080)) & 0xff00ff00;

		return px_rb + px_ag;
	} else {
		int iu = (int)u;
		int iv = (int)v;

		if (mbClampU && (unsigned)iu > mULimit)
			iu = ~(iu >> 31) & mULimit;
		if (mbClampV && (unsigned)iv > mVLimit)
			iv = ~(iv >> 31) & mVLimit;

		return ((const uint32 *)((const char *)mpSrc + mPitch*iv))[iu];
	}
}

///////////////////////////////////////////////////////////////////////////////

VDShaderInterpreter::VDShaderInterpreter() {
}

VDShaderInterpreter::~VDShaderInterpreter() {
}

void VDShaderInterpreter::LoadSamplers(const VDShaderInterpreterSampler *samplers, int numSamplers) {
	mpSamplers = samplers;
}

bool VDShaderInterpreter::LoadShader(const VDShaderFunction& func, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	mScalarForm.Clear();

	VDShaderScalarTransformer xform(mScalarForm);
	if (!xform.LoadShader(func, pElements, numElements, out))
		return false;

	int numConstants = 0;
	int numTemps = 0;
	int numOutputs = 0;
	int numInputs = mScalarForm.mInputCount;

	for(VDShaderScalarForm::Insns::const_iterator it(mScalarForm.mInsns.begin()), itEnd(mScalarForm.mInsns.end()); it != itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		const int registers[] = { insn.dst, insn.src1, insn.src2, insn.src3 };

		for(int i=0; i<sizeof(registers)/sizeof(registers[0]); ++i) {
			int r = registers[i];

			if (r >= 0) {
				int idx = r & kVDShaderRegIndexMask;

				switch(r & kVDShaderRegTypeMask) {
					case kVDShaderRegR0:
						if (idx >= numTemps)
							numTemps = idx + 1;
						break;

					case kVDShaderRegC0:
						if (idx >= numConstants)
							numConstants = idx + 1;
						break;

					case kVDShaderRegO0:
						if (idx >= numOutputs)
							numOutputs = idx + 1;
						break;
				}
			}
		}
	}

	mInputCount = numInputs;

	const int constBase = 0;
	const int inputBase = constBase + numConstants;
	const int tempBase = inputBase + numInputs;
	const int outputBase = tempBase + numTemps;

	for(VDShaderScalarForm::Insns::iterator it(mScalarForm.mInsns.begin()), itEnd(mScalarForm.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		int registers[] = { insn.dst, insn.src1, insn.src2, insn.src3 };

		for(int i=0; i<sizeof(registers)/sizeof(registers[0]); ++i) {
			int& r = registers[i];

			if (r) {
				int idx = r & kVDShaderRegIndexMask;

				switch(r & kVDShaderRegTypeMask) {
					case kVDShaderRegR0:
						r = tempBase + idx;
						break;

					case kVDShaderRegC0:
						r = constBase + idx;
						break;

					case kVDShaderRegO0:
						r = outputBase + idx;
						break;

					case kVDShaderRegV0:
						r = inputBase + idx;
						break;
				}
			}
		}

		insn.dst = registers[0];
		insn.src1 = registers[1];
		insn.src2 = registers[2];
		insn.src3 = registers[3];
	}

	VDShaderInterpreterTransformedInsn end = {0};
	mScalarForm.mInsns.push_back(end);

	int slotCount = numInputs + numTemps + numOutputs + numConstants;

	mInputBase = inputBase;
	mOutputBase = outputBase;

	mRegisterFile.clear();
	mRegisterFile.reserve(slotCount);
	mRegisterFile = mScalarForm.mConstants;
	mRegisterFile.resize(slotCount, 0);

	return true;
}

void VDShaderInterpreter::Execute(uint32 *dst, const VDShaderInterpreterInterpolator *interps, uint32 count) {
	const VDShaderScalarForm::TransformedInsn * __restrict ip0 = &mScalarForm.mInsns[0];
	float * const r = mRegisterFile.data();
	float * const in = &r[mInputBase];
	const float * const out = &r[mOutputBase];

	for(int i=0; i<mInputCount; ++i) {
		in[i*2+0] = interps[0].u[i];
		in[i*2+1] = interps[0].dudx[i];
	}

	while(count--) {
		const VDShaderScalarForm::TransformedInsn * __restrict ip = ip0;

		for(;;) {
			int op = ip->op;
			if (!op)
				break;

#define R(index) r[ip->index]

			switch(ip->op) {
				case kVDShaderInterpOp_beginloop:
					ip0 = ip + 1;
					break;

				case kVDShaderInterpOp_mov:
					R(dst) = R(src1);
					break;

				case kVDShaderInterpOp_add:
					R(dst) = R(src1) + R(src2);
					break;

				case kVDShaderInterpOp_sub:
					R(dst) = R(src1) - R(src2);
					break;

				case kVDShaderInterpOp_mul:
					R(dst) = R(src1) * R(src2);
					break;

				case kVDShaderInterpOp_min:
					R(dst) = R(src1) < R(src2) ? R(src1) : R(src2);
					break;

				case kVDShaderInterpOp_max:
					R(dst) = R(src1) > R(src2) ? R(src1) : R(src2);
					break;

				case kVDShaderInterpOp_rsq:
					R(dst) = 1.0f / sqrtf(fabsf(R(src1)));
					break;

				case kVDShaderInterpOp_rcp:
					R(dst) = 1.0f / R(src1);
					break;

				case kVDShaderInterpOp_frc:
					R(dst) = R(src1) - floorf(R(src1));
					break;

				case kVDShaderInterpOp_neg:
					R(dst) = -R(src1);
					break;

				case kVDShaderInterpOp_abs:
					R(dst) = fabsf(R(src1));
					break;

				case kVDShaderInterpOp_log:
					{
						float v = fabsf(R(src1));

						if (v == 0.0f)
							v = -FLT_MAX;
						else
							v = logf(v) * 1.4426950408889634073599246810019f;

						R(dst) = v;
					}
					break;

				case kVDShaderInterpOp_exp:
					R(dst) = expf(R(src1) * 0.69314718055994530941723212145818f);
					break;

				case kVDShaderInterpOp_cmp:
					R(dst) = (R(src1) >= 0) ? R(src2) : R(src3);
					break;

				case kVDShaderInterpOp_texld:
					(uint32&)R(dst) = mpSamplers[0].Sample(R(src1), R(src2));
					break;

				case kVDShaderInterpOp_unpackr:
					R(dst) = (float)(((uint32&)R(src1) >> 16) & 0xff) * (1.0f / 255.0f);
					break;

				case kVDShaderInterpOp_unpackg:
					R(dst) = (float)(((uint32&)R(src1) >>  8) & 0xff) * (1.0f / 255.0f);
					break;

				case kVDShaderInterpOp_unpackb:
					R(dst) = (float)(((uint32&)R(src1)      ) & 0xff) * (1.0f / 255.0f);
					break;

				case kVDShaderInterpOp_unpacka:
					R(dst) = (float)(((uint32&)R(src1) >> 24) & 0xff) * (1.0f / 255.0f);
					break;

			}

#undef R

			++ip;
		}

		union {
			float f;
			sint32 i;
		}	convr = { out[0] * 255.0f + 12582912.0f },
			convg = { out[1] * 255.0f + 12582912.0f },
			convb = { out[2] * 255.0f + 12582912.0f },
			conva = { out[3] * 255.0f + 12582912.0f };

		sint32 r = convr.i < 0x4b400000 ? 0 : convr.i - 0x4b400000;
		sint32 g = convg.i < 0x4b400000 ? 0 : convg.i - 0x4b400000;
		sint32 b = convb.i < 0x4b400000 ? 0 : convb.i - 0x4b400000;
		sint32 a = conva.i < 0x4b400000 ? 0 : conva.i - 0x4b400000;

		r += (255 - r) & ((255 - r) >> 31);
		g += (255 - g) & ((255 - g) >> 31);
		b += (255 - b) & ((255 - b) >> 31);
		a += (255 - a) & ((255 - a) >> 31);

		*dst++ = (a << 24) + (r << 16) + (g << 8) + b;
	}
}

