//	VDShader - Custom shader video filter for VirtualDub
//	Copyright (C) 2007 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "lexer.h"
#include "log.h"

using namespace nsVDShaderTokens;

void VDShaderWriteLogOutputF(IVDShaderLogOutput& out, const char *format ...) {
	char buf[3072];
	va_list val;

	va_start(val, format);
	int len = _vsnprintf(buf, 3071, format, val);
	va_end(val);

	if ((unsigned)len <= 3071) {
		buf[len] = 0;
		out.WriteLogOutput(buf);
	}
}

void VDShaderLexer::Init(const char *src, size_t len, const char *context, IVDShaderLogOutput *pOutput) {
	mpSrc = src;
	mpSrcEnd = src + len;
	mpSrcLineStart = src;
	mpToken = src;
	mpTokenLineStart = src;
	mpContext = context;
	mLineNo = 1;
	mTokenLength = 0;
	mbError = false;
	mpOutput = pOutput;
}

bool VDShaderLexer::GetErrorInfo(VDShaderErrorInfo& errorInfo) {
	if (!mbError)
		return false;

	errorInfo = mErrorInfo;
	return true;
}

int VDShaderLexer::Token() {
	bool inComment = false;
	char c;

	// parse out whitespace and comments
	for(;;) {
		if (mpSrc == mpSrcEnd) {
			mpTokenLineStart = mpSrcLineStart;
			mpToken = mpSrc;
			mTokenLength = 0;
			return 0;
		}

		c = *mpSrc++;
		if (c == ' ' || c == '\t')
			continue;

		if (c == '\n' || c == '\r') {
			if (mpSrc != mpSrcEnd) {
				char d = *mpSrc;
				if ((c ^ d) == ('\n' ^ '\r'))
					++mpSrc;
			}
			++mLineNo;
			mpSrcLineStart = mpSrc;
			continue;
		}

		if (inComment) {
			if (c == '*' && mpSrc != mpSrcEnd && *mpSrc == '/') {
				++mpSrc;

				inComment = false;
			}
		} else {
			if (c == '/' && mpSrc != mpSrcEnd) {
				char d = *mpSrc;
				if (d == '/') {				// C++-style comment
					++mpSrc;
					while(mpSrc != mpSrcEnd) {
						c = *mpSrc;

						if (c == '\n' || c == '\r')
							break;

						++mpSrc;
					}
					continue;
				} else if (d == '*') {		// C-style comment
					++mpSrc;

					inComment = true;
					continue;
				}
			}

			break;
		}
	}

	mpToken = mpSrc - 1;
	mpTokenLineStart = mpSrcLineStart;

	// check for numeric tokens
	if ((unsigned char)(c - '0') < 10) {
		bool overflow = false;
		int value = c - '0';

		while(mpSrc != mpSrcEnd) {
			c = *mpSrc;

			if (c == '.') {
				double fvalue = 0;

				for(const char *s = mpToken; s != mpSrc; ++s)
					fvalue = fvalue * 10.0 + (float)(*s - '0');

				++mpSrc;

				double scale = 0.1;
				while(mpSrc != mpSrcEnd) {
					c = *mpSrc;

					int digit = (c - '0');
					if ((unsigned)digit >= 10)
						break;

					++mpSrc;
					fvalue += scale * (float)digit;
					scale *= 0.1;
				}

				if (mpSrc != mpSrcEnd) {
					c = *mpSrc;
					if (c == 'e' || c == 'E') {
						++mpSrc;

						const char *pExpBase = mpSrc;
						int exp = 0;

						while(mpSrc != mpSrcEnd) {
							c = *mpSrc;

							int digit = (c - '0');
							if ((unsigned)digit >= 10)
								break;

							exp = (exp*10) + digit;

							++mpSrc;
						}

						if (mpSrc == pExpBase) {
							SetError("Invalid real number");
							return 0;
						}

						fvalue *= pow(10.0, exp);
					}
				}

				if (mpSrc != mpSrcEnd) {
					c = *mpSrc;
					if (c == 'f' || c == 'F')
						++mpSrc;
				}

				mFloatVal = (float)fvalue;
				mTokenLength = mpSrc - mpToken;
				return kTokenFloat;
			}

			int digit = (c - '0');
			if ((unsigned)digit >= 10)
				break;
			++mpSrc;

			if (value > 214748364)
				overflow = true;

			value *= 10;

			if (value > 2147483647 - digit)
				overflow = true;

			value += digit;
		
		}

		if (overflow) {
			SetError("Integer literal too large");
			return 0;
		}

		mIntVal = value;
		mTokenLength = mpSrc - mpToken;
		return kTokenInt;
	}

	// check for identifiers
	if (c == '_' || (unsigned char)((c & 0xdf) - 'A') < 26) {
		while(mpSrc != mpSrcEnd) {
			c = *mpSrc++;

			if (c != '_' && (unsigned char)((c & 0xdf) - 'A') >= 26 && (unsigned char)(c - '0') >= 10) {
				--mpSrc;
				break;
			}
		}

		mTokenLength = mpSrc - mpToken;
		return kTokenIdent;
	}

	// check for compound tokens
	if (mpSrcEnd != mpSrc) {
		const char d = *mpSrc;

		if (c == '!') {
			if (d == '=') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenNE;
			}
		} else if (c == '=') {
			if (d == '=') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenEQ;
			}
		} else if (c == '<') {
			if (d == '=') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenLE;
			}
		} else if (c == '>') {
			if (d == '=') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenGE;
			}
		} else if (c == '&') {
			if (d == '&') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenAnd;
			}
		} else if (c == '|') {
			if (d == '|') {
				mTokenLength = 2;
				++mpSrc;
				return kTokenOr;
			}
		}
	}

	if ((unsigned char)(c - 0x20) >= 0x5f) {
		SetErrorF("Unrecognized character '%c'", c);
		return 0;
	}

	mTokenLength = 1;
	return c;
}

void VDShaderLexer::Push() {
	mpSrc = mpToken;
}

bool VDShaderLexer::SetError(const char *s) {
	if (mbError)
		return false;

	mbError = true;
	mErrorInfo.mLine = mLineNo;
	mErrorInfo.mColumn = (mpToken - mpSrcLineStart) + 1;
	mErrorInfo.mLength = mTokenLength;

	char buf[2048];
	if (_snprintf(buf, 2047, "%s(%d,%d): Error! %s\r\n", mpContext, mLineNo, (int)(mpSrc - mpSrcLineStart) + 1, s) > 0) {
		buf[2047] = 0;
		mpOutput->WriteLogOutput(buf);
	}
	return false;
}

bool VDShaderLexer::SetErrorF(const char *format, ...) {
	if (mbError)
		return false;

	mbError = true;
	mErrorInfo.mLine = mLineNo;
	mErrorInfo.mColumn = (mpToken - mpSrcLineStart) + 1;
	mErrorInfo.mLength = mTokenLength;

	va_list val;
	char buf[2048], *s = buf, *limit = s + 2047;
	int len;
	len = _snprintf(s, limit-s, "%s(%d,%d): Error! ", mpContext, mLineNo, (int)(mpSrc - mpSrcLineStart) + 1);
	if (len >= 0) {
		s += len;
		va_start(val, format);
		len = _vsnprintf(s, limit-s, format, val);
		va_end(val);

		if (len >= 0) {
			s += len;

			if (limit-s >= 2) {
				*s++ = '\r';
				*s++ = '\n';
				*s = 0;
				mpOutput->WriteLogOutput(buf);
			}
		}
	}
	return false;
}
