#include "stdafx.h"
#include <vector>
#include <float.h>
#include "parser.h"
#include "log.h"
#include "disasm.h"
#include "scalar.h"

void VDShaderTransformFunctionSimplify(VDShaderFunction& func);

///////////////////////////////////////////////////////////////////////////////
void Disassemble(IVDShaderLogOutput &out, const VDShaderInterpreterTransformedInsn& insn) {
	switch(insn.op) {
		case kVDShaderInterpOp_end:
			out.WriteLogOutput("end");
			break;
		case kVDShaderInterpOp_mov:
			out.WriteLogOutput("mov");
			break;
		case kVDShaderInterpOp_add:
			out.WriteLogOutput("add");
			break;
		case kVDShaderInterpOp_sub:
			out.WriteLogOutput("sub");
			break;
		case kVDShaderInterpOp_mul:
			out.WriteLogOutput("mul");
			break;
		case kVDShaderInterpOp_rsq:
			out.WriteLogOutput("rsq");
			break;
		case kVDShaderInterpOp_rcp:
			out.WriteLogOutput("rcp");
			break;
		case kVDShaderInterpOp_frc:
			out.WriteLogOutput("frc");
			break;
		case kVDShaderInterpOp_neg:
			out.WriteLogOutput("neg");
			break;
		case kVDShaderInterpOp_min:
			out.WriteLogOutput("min");
			break;
		case kVDShaderInterpOp_max:
			out.WriteLogOutput("max");
			break;
		case kVDShaderInterpOp_abs:
			out.WriteLogOutput("abs");
			break;
		case kVDShaderInterpOp_log:
			out.WriteLogOutput("log");
			break;
		case kVDShaderInterpOp_exp:
			out.WriteLogOutput("exp");
			break;
		case kVDShaderInterpOp_cmp:
			out.WriteLogOutput("cmp");
			break;
		case kVDShaderInterpOp_texld:
			out.WriteLogOutput("texld");
			break;
		case kVDShaderInterpOp_unpackr:
			out.WriteLogOutput("unpackr");
			break;
		case kVDShaderInterpOp_unpackg:
			out.WriteLogOutput("unpackg");
			break;
		case kVDShaderInterpOp_unpackb:
			out.WriteLogOutput("unpackb");
			break;
		case kVDShaderInterpOp_unpacka:
			out.WriteLogOutput("unpacka");
			break;
		case kVDShaderInterpOp_beginloop:
			out.WriteLogOutput("beginloop");
			break;
		default:
			VDASSERT(false);
	}

	if (insn.dst) {
		switch(insn.dst & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, " r%d", insn.dst - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, " c%d", insn.dst - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, " o%d", insn.dst - kVDShaderRegO0);
				break;

			default:
				out.WriteLogOutput(" <invalid>");
				break;
		}
	}

	if (insn.src1) {
		if (insn.dst)
			out.WriteLogOutput(",");

		switch(insn.src1 & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, " r%d", insn.src1 - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, " c%d", insn.src1 - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, " o%d", insn.src1 - kVDShaderRegO0);
				break;

			case kVDShaderRegV0:
				VDShaderWriteLogOutputF(out, " v%d", insn.src1 - kVDShaderRegV0);
				break;

			default:
				out.WriteLogOutput(" <invalid>");
				break;
		}
	}

	if (insn.src2) {
		switch(insn.src2 & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, ", r%d", insn.src2 - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, ", c%d", insn.src2 - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, ", o%d", insn.src2 - kVDShaderRegO0);
				break;

			case kVDShaderRegV0:
				VDShaderWriteLogOutputF(out, ", v%d", insn.src2 - kVDShaderRegV0);
				break;

			default:
				out.WriteLogOutput(", <invalid>");
				break;
		}
	}

	if (insn.src3) {
		switch(insn.src3 & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				VDShaderWriteLogOutputF(out, ", r%d", insn.src3 - kVDShaderRegR0);
				break;

			case kVDShaderRegC0:
				VDShaderWriteLogOutputF(out, ", c%d", insn.src3 - kVDShaderRegC0);
				break;

			case kVDShaderRegO0:
				VDShaderWriteLogOutputF(out, ", o%d", insn.src3 - kVDShaderRegO0);
				break;

			case kVDShaderRegV0:
				VDShaderWriteLogOutputF(out, ", v%d", insn.src3 - kVDShaderRegV0);
				break;

			default:
				out.WriteLogOutput(", <invalid>");
				break;
		}
	}

	out.WriteLogOutput("\n");
}

void Disassemble(IVDShaderLogOutput& out, const VDShaderScalarForm& shader) {
	const int numConstants = shader.mConstants.size();

	for(int i=0; i<numConstants; ++i)
		VDShaderWriteLogOutputF(out, "def c%d, %g\n", i, shader.mConstants[i]);

	for(vdfastvector<VDShaderInterpreterTransformedInsn>::const_iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		Disassemble(out, insn);
	}
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarApplyCopyPropagation(VDShaderScalarForm& shader) {
	vdfastvector<int> tempRemap(shader.mTempCount);
	vdfastvector<bool> tempWrittenInLoop(shader.mTempCount, false);

	for(int i=0; i<shader.mTempCount; ++i)
		tempRemap[i] = kVDShaderRegR0 + i;

	// Pre-iterate through the shader and find all registers that are
	// written in the loop.
	bool inLoop = false;
	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op == kVDShaderInterpOp_beginloop)
			inLoop = true;

		if (inLoop && (insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
			tempWrittenInLoop[insn.dst - kVDShaderRegR0] = true;
	}

	// Apply copy propagation.
	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;

		if (insn.op == kVDShaderInterpOp_beginloop) {
			// At the beginning of the loop, any register that we attempt to read
			// that was written at the end of the loop is a loop carried dependency
			// that we cannot remap, as it is written by both the prologue and in
			// the loop.
			for(int i=0; i<shader.mTempCount; ++i) {
				if (tempWrittenInLoop[i])
					tempRemap[i] = kVDShaderRegR0 + i;
			}
		}

		// remap source registers
		if ((insn.src1 & kVDShaderRegTypeMask) == kVDShaderRegR0)
			insn.src1 = tempRemap[insn.src1 - kVDShaderRegR0];

		if ((insn.src2 & kVDShaderRegTypeMask) == kVDShaderRegR0)
			insn.src2 = tempRemap[insn.src2 - kVDShaderRegR0];

		if ((insn.src3 & kVDShaderRegTypeMask) == kVDShaderRegR0)
			insn.src3 = tempRemap[insn.src3 - kVDShaderRegR0];

		// map dest register if this is a mov insn
		if (insn.op == kVDShaderInterpOp_mov) {
			if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
				tempRemap[insn.dst - kVDShaderRegR0] = insn.src1;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarApplyDeadStoreElimination(VDShaderScalarForm& shader) {
	int numInsns = shader.mInsns.size();

	vdfastvector<int> regWriteHead(shader.mTempCount, -1);
	vdfastvector<int> regWriteNext(numInsns, -1);
	vdfastvector<int> insnQueue;
	vdfastvector<bool> markedInsns(numInsns, false);

	// Iterate forwards through the instruction stream construct a list of
	// (a) instructions that write each temporary, and (b) instructions that
	// write outputs.
	for(int i=0; i<numInsns; ++i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (!insn.dst) {
			insnQueue.push_back(i);
			markedInsns[i] = true;
		}

		int dstidx = insn.dst & kVDShaderRegIndexMask;
		switch(insn.dst & kVDShaderRegTypeMask) {
			case kVDShaderRegR0:
				regWriteNext[i] = regWriteHead[dstidx];
				regWriteHead[dstidx] = i;
				break;
			case kVDShaderRegO0:
				insnQueue.push_back(i);
				markedInsns[i] = true;
				break;
		}
	}

	// Follow reference graph and mark all instructions that 
	// Iterate backwards through the instruction stream and eliminate all
	// writes that aren't referenced.
	while(!insnQueue.empty()) {
		int i = insnQueue.back();
		insnQueue.pop_back();

		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		// queue instructions that write sources
		const int sources[] = {insn.src1, insn.src2, insn.src3};

		for(int j=0; j<sizeof(sources)/sizeof(sources[0]); ++j) {
			int r = sources[j];

			if ((r & kVDShaderRegTypeMask) == kVDShaderRegR0) {
				int ridx = r - kVDShaderRegR0;

				for(int k = regWriteHead[ridx]; k >= 0; k = regWriteNext[k]) {
					if (!markedInsns[k]) {
						markedInsns[k] = true;
						insnQueue.push_back(k);
					}
				}
			}
		}
	}

	// Delete all unmarked instructions.
	VDShaderScalarForm::Insns::iterator itDst(shader.mInsns.begin());
	for(int i=0; i<numInsns; ++i) {
		const VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (markedInsns[i]) {
			if (&*itDst != &insn)
				*itDst = insn;

			++itDst;
		}
	}

	shader.mInsns.erase(itDst, shader.mInsns.end());
}

///////////////////////////////////////////////////////////////////////////////

void VDShaderScalarPackConstants(VDShaderScalarForm& shader) {
	vdfastvector<int> constantMap(shader.mConstants.size(), -1);
	vdfastvector<float> newConstants;

	for(VDShaderScalarForm::Insns::iterator it(shader.mInsns.begin()), itEnd(shader.mInsns.end()); it != itEnd; ++it) {
		VDShaderInterpreterTransformedInsn& insn = *it;
		int sources[3] = { insn.src1, insn.src2, insn.src3 };

		for(int i=0; i<3; ++i) {
			int r = sources[i];

			if ((r & kVDShaderRegTypeMask) == kVDShaderRegC0) {
				int idx = r - kVDShaderRegC0;
				int newIdx = constantMap[idx];

				if (newIdx < 0) {
					newIdx = newConstants.size();
					newConstants.push_back(shader.mConstants[idx]);
					constantMap[idx] = newIdx;
				}

				sources[i] = newIdx + kVDShaderRegC0;
			}
		}

		insn.src1 = sources[0];
		insn.src2 = sources[1];
		insn.src3 = sources[2];
	}

	shader.mConstants.swap(newConstants);
}

bool VDShaderScalarPackTemps(VDShaderScalarForm& shader, IVDShaderLogOutput *out) {
	const int numInsns = shader.mInsns.size();

	vdfastvector<int> tempMap(shader.mTempCount, -1);
	vdfastvector<bool> tempWritten(shader.mTempCount, false);
	int newTempCount = 0;

	// Iterate forwards through the instruction stream and mark all temp
	// reads that aren't hit by a write. These are loop carried dependencies
	// and must be considered live at the end of the loop.
	bool inLoop = false;
	for(int i=0; i<numInsns; ++i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		if (insn.op == kVDShaderInterpOp_beginloop)
			inLoop = true;
		else if (!inLoop)
			continue;

		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0)
			tempWritten[insn.dst - kVDShaderRegR0] = true;

		if ((insn.src1 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int idx1 = insn.src1 - kVDShaderRegR0;

			if (!tempWritten[idx1] && tempMap[idx1] < 0)
				tempMap[idx1] = newTempCount++;
		}

		if ((insn.src2 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int idx2 = insn.src2 - kVDShaderRegR0;

			if (!tempWritten[idx2] && tempMap[idx2] < 0)
				tempMap[idx2] = newTempCount++;
		}

		if ((insn.src3 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int idx3 = insn.src3 - kVDShaderRegR0;

			if (!tempWritten[idx3] && tempMap[idx3] < 0)
				tempMap[idx3] = newTempCount++;
		}
	}

	vdfastvector<int> freeTemps;

	for(int i=numInsns - 1; i >= 0; --i) {
		VDShaderInterpreterTransformedInsn& insn = shader.mInsns[i];

		// remap and kill destination register
		if ((insn.dst & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int dstidx = insn.dst - kVDShaderRegR0;
			int dstidx2 = tempMap[dstidx];

			if (dstidx2 < 0) {
				if (out) {
					out->WriteLogOutput("Internal Compiler Error: Dead store detected during register allocation pass.\n");
					out->WriteLogOutput("  Bad instruction:");
					Disassemble(*out, insn);
				}
				return false;
			}

			insn.dst = kVDShaderRegR0 + dstidx2;

			// We shouldn't kill the destination register if it's the same as one of the sources. Not
			// only is it unnecessary, but it can also break loop carried dependencies.
			if (insn.dst != insn.src1 && insn.dst != insn.src2 && insn.dst != insn.src3) {
				freeTemps.push_back(dstidx2);
				tempMap[dstidx] = -1;
			}
		}

		// allocate and remap source register 1
		if ((insn.src1 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int src1Idx = insn.src1 - kVDShaderRegR0;
			int src1Idx2 = tempMap[src1Idx];

			if (src1Idx2 < 0) {
				if (freeTemps.empty())
					freeTemps.push_back(newTempCount++);

				src1Idx2 = freeTemps.back();
				freeTemps.pop_back();

				tempMap[src1Idx] = src1Idx2;
			}

			insn.src1 = kVDShaderRegR0 + src1Idx2;
		}

		// allocate and remap source register 2
		if ((insn.src2 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int src2Idx = insn.src2 - kVDShaderRegR0;
			int src2Idx2 = tempMap[src2Idx];

			if (src2Idx2 < 0) {
				if (freeTemps.empty())
					freeTemps.push_back(newTempCount++);

				src2Idx2 = freeTemps.back();
				freeTemps.pop_back();

				tempMap[src2Idx] = src2Idx2;
			}

			insn.src2 = kVDShaderRegR0 + src2Idx2;
		}

		// allocate and remap source register 3
		if ((insn.src3 & kVDShaderRegTypeMask) == kVDShaderRegR0) {
			int src3Idx = insn.src3 - kVDShaderRegR0;
			int src3Idx2 = tempMap[src3Idx];

			if (src3Idx2 < 0) {
				if (freeTemps.empty())
					freeTemps.push_back(newTempCount++);

				src3Idx2 = freeTemps.back();
				freeTemps.pop_back();

				tempMap[src3Idx] = src3Idx2;
			}

			insn.src3 = kVDShaderRegR0 + src3Idx2;
		}
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderScalarForm::VDShaderScalarForm() {
}

VDShaderScalarForm::~VDShaderScalarForm() {
}

void VDShaderScalarForm::Clear() {
	mInsns.clear();
	mConstants.clear();
	mTempCount = 0;
}

///////////////////////////////////////////////////////////////////////////////

VDShaderScalarTransformer::VDShaderScalarTransformer(VDShaderScalarForm& dst)
	: mShader(dst)
{
}

VDShaderScalarTransformer::~VDShaderScalarTransformer() {
}

bool VDShaderScalarTransformer::LoadShader(const VDShaderFunction& func0, const VDShaderBindingElement *pElements, int numElements, IVDShaderLogOutput *out) {
	VDShaderFunction func(func0);

	// transform complex vector instructions to simpler ones
	VDShaderTransformFunctionSimplify(func);

	if (out)
		VDShaderDisassembleFunction(*out, func, NULL);

	// load constants
	int numSrcConstants = func.mConstants.size() * 4;
	const float *srcConstants = (const float *)func.mConstants.data();
	mShader.mConstants.assign(srcConstants, srcConstants + numSrcConstants);

	// transform instructions
	int nextDstReg = kVDShaderRegR0 + func.mRegisterCount*4;

	int t0_x = nextDstReg + 0;
	int t0_y = nextDstReg + 1;
	nextDstReg += 4;

	mov(t0_x, kVDShaderRegV0);
	mov(t0_y, kVDShaderRegV0 + 2);

	beginloop();

	for(VDShaderFunction::Insns::const_iterator it(func.mInsns.begin()), itEnd(func.mInsns.end()); it!=itEnd; ++it) {
		const VDShaderInsn& insn = *it;

		struct DecodedInsn {
			int op;
			int dst[4];
			int src[16];
		};

		DecodedInsn dinsn;

		for(int i=0; i<4; ++i) {
			const VDShaderInsnSource& src = insn.mSources[i];

			if (!src.mRegister)
				continue;

			int srcRegIndex = src.mRegister & kVDShaderRegIndexMask;
			int srcBase = i*4;

			switch(src.mRegister & kVDShaderRegTypeMask) {
				case kVDShaderRegR0:
					dinsn.src[srcBase+0] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = kVDShaderRegR0 + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				case kVDShaderRegC0:
					dinsn.src[srcBase+0] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = kVDShaderRegC0 + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				case kVDShaderRegS0:
					break;

				case kVDShaderRegV0:
					dinsn.src[srcBase+0] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 0) & 3);
					dinsn.src[srcBase+1] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 2) & 3);
					dinsn.src[srcBase+2] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 4) & 3);
					dinsn.src[srcBase+3] = t0_x + srcRegIndex*4 + ((src.mSwizzle >> 6) & 3);
					break;

				default:
					VDASSERT(false);
					break;
			}

			if (src.mModifiers & kVDShaderMod_Negate) {
				for(int j=0; j<4; ++j) {
					neg(nextDstReg, dinsn.src[srcBase+j]);
					dinsn.src[srcBase+j] = nextDstReg++;
				}
			}
		}

		int dstReg = insn.mDestRegister;
		int dstRegIndex = dstReg & kVDShaderRegIndexMask;
		int writeMask = insn.mWriteMask;

		switch(dstReg & kVDShaderRegTypeMask) {
			case kVDShaderRegO0:
				dstRegIndex = kVDShaderRegO0 + dstRegIndex*4;
				break;

			case kVDShaderRegR0:
				dstRegIndex = kVDShaderRegR0 + dstRegIndex*4;
				break;

			default:
				VDASSERT(false);
				break;
		}

		for(int i=0; i<4; ++i) {
			if (writeMask & (1 << i))
				dinsn.dst[i] = dstRegIndex + i;
			else
				dinsn.dst[i] = 0;
		}

		switch(insn.mOp) {
			case kVDShaderOp_mov:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						mov(dinsn.dst[i], dinsn.src[i]);
				}
				break;
			case kVDShaderOp_add:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						add(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_cmp:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						cmp(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4], dinsn.src[i+8]);
				}
				break;
			case kVDShaderOp_sub:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						sub(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mul:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						mul(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_mad:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i]) {
						int t = nextDstReg++;
						mul(t, dinsn.src[i], dinsn.src[i+4]);
						add(dinsn.dst[i], t, dinsn.src[i+8]);
					}
				}
				break;
			case kVDShaderOp_min:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						min(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_max:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						max(dinsn.dst[i], dinsn.src[i], dinsn.src[i+4]);
				}
				break;
			case kVDShaderOp_dp3:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t5) {
								t5 = dinsn.dst[i];

								mul(t, dinsn.src[0], dinsn.src[4]);
								mul(t2, dinsn.src[1], dinsn.src[5]);
								add(t3, t, t2);
								mul(t4, dinsn.src[2], dinsn.src[6]);
								add(t5, t3, t4);
							} else
								mov(dinsn.dst[i], t5);
						}
					}
				}
				break;
			case kVDShaderOp_dp4:
				{
					int t = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;
					int t4 = nextDstReg++;
					int t5 = nextDstReg++;
					int t6 = nextDstReg++;
					int t7 = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t7) {
								t7 = dinsn.dst[i];

								mul(t, dinsn.src[0], dinsn.src[4]);
								mul(t2, dinsn.src[1], dinsn.src[5]);
								add(t3, t, t2);
								mul(t4, dinsn.src[2], dinsn.src[6]);
								add(t5, t3, t4);
								mul(t6, dinsn.src[3], dinsn.src[7]);
								add(t7, t5, t6);
							} else
								mov(dinsn.dst[i], t7);
						}
					}
				}
				break;
			case kVDShaderOp_rsq:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								rsq(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_rcp:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								rcp(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;
			case kVDShaderOp_frc:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								frc(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_abs:
				for(int i=0; i<4; ++i) {
					if (dinsn.dst[i])
						abs(dinsn.dst[i], dinsn.src[i]);
				}
				break;

			case kVDShaderOp_crs:
				// crs r0, r1, r2 becomes:
				//
				//		mul r0, r1.zxy, r2.yzx
				//		mad r0, r1.yzx, r2.zxy, -r0

				if (dinsn.dst[0]) {
					mul(nextDstReg+0, dinsn.src[1], dinsn.src[6]);
					mul(nextDstReg+1, dinsn.src[2], dinsn.src[5]);
					sub(dinsn.dst[0], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				if (dinsn.dst[1]) {
					mul(nextDstReg+0, dinsn.src[2], dinsn.src[4]);
					mul(nextDstReg+1, dinsn.src[0], dinsn.src[6]);
					sub(dinsn.dst[1], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				if (dinsn.dst[2]) {
					mul(nextDstReg+0, dinsn.src[0], dinsn.src[5]);
					mul(nextDstReg+1, dinsn.src[1], dinsn.src[4]);
					sub(dinsn.dst[2], nextDstReg+0, nextDstReg+1);
					nextDstReg += 2;
				}

				break;

			case kVDShaderOp_dp2add:
				{
					int t0 = nextDstReg++;
					int t1 = nextDstReg++;
					int t2 = nextDstReg++;
					int t3 = nextDstReg++;

					mul(t0, dinsn.src[0], dinsn.src[4]);
					mul(t1, dinsn.src[1], dinsn.src[5]);
					add(t2, t0, t1);
					add(t3, t2, dinsn.src[8]);

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i])
							mov(dinsn.dst[i], t3);
					}
				}
				break;

			case kVDShaderOp_log:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								log(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_exp:
				{
					int t = 0;

					for(int i=0; i<4; ++i) {
						if (dinsn.dst[i]) {
							if (!t) {
								t = dinsn.dst[i];

								exp(t, dinsn.src[0]);
							} else
								mov(dinsn.dst[i], t);
						}
					}
				}
				break;

			case kVDShaderOp_texld:
				{
					int t = nextDstReg++;
					texld(t, dinsn.src[0], dinsn.src[1]);

					if (dinsn.dst[0])
						unpackr(dinsn.dst[0], t);

					if (dinsn.dst[1])
						unpackg(dinsn.dst[1], t);

					if (dinsn.dst[2])
						unpackb(dinsn.dst[2], t);

					if (dinsn.dst[3])
						unpacka(dinsn.dst[3], t);
				}
				break;

			default:
				VDASSERT(false);
		}
	}

	// update interpolators
	add(t0_x, t0_x, kVDShaderRegV0 + 1);
	add(t0_y, t0_y, kVDShaderRegV0 + 3);

	mShader.mTempCount = nextDstReg - kVDShaderRegR0;

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Initial conversion result:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// apply copy propagation
	VDShaderScalarApplyCopyPropagation(mShader);

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Copy propagation output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// apply dead store elimination
	VDShaderScalarApplyDeadStoreElimination(mShader);

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Dead store elimination output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// eliminate redundant and unused constants
	VDShaderScalarPackConstants(mShader);

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Constant pack output:\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	// pack temporary registers
	if (!VDShaderScalarPackTemps(mShader, out))
		return false;

	if (out) {
		out->WriteLogOutput("//\n");
		out->WriteLogOutput("// Generated by VirtualDub vector-to-scalar recompiler\n");
		out->WriteLogOutput("//\n");

		Disassemble(*out, mShader);
	}

	int numConstants = 0;
	int numTemps = 0;
	int numOutputs = 0;
	int numInputs = 0;

	for(VDShaderScalarForm::Insns::const_iterator it(mShader.mInsns.begin()), itEnd(mShader.mInsns.end()); it != itEnd; ++it) {
		const VDShaderInterpreterTransformedInsn& insn = *it;

		const int registers[] = { insn.dst, insn.src1, insn.src2, insn.src3 };

		for(int i=0; i<sizeof(registers)/sizeof(registers[0]); ++i) {
			int r = registers[i];

			if (r >= 0) {
				int idx = r & kVDShaderRegIndexMask;

				switch(r & kVDShaderRegTypeMask) {
					case kVDShaderRegR0:
						if (idx >= numTemps)
							numTemps = idx + 1;
						break;

					case kVDShaderRegC0:
						if (idx >= numConstants)
							numConstants = idx + 1;
						break;

					case kVDShaderRegO0:
						if (idx >= numOutputs)
							numOutputs = idx + 1;
						break;

					case kVDShaderRegV0:
						if (idx >= numInputs)
							numInputs = idx + 1;
						break;
				}
			}
		}
	}

	mShader.mInputCount = numInputs;

	return true;
}

void VDShaderScalarTransformer::mov(int dst, int src) {
	push_insn(kVDShaderInterpOp_mov, dst, src, 0);
}

void VDShaderScalarTransformer::add(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_add, dst, src1, src2);
}

void VDShaderScalarTransformer::sub(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_sub, dst, src1, src2);
}

void VDShaderScalarTransformer::mul(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_mul, dst, src1, src2);
}

void VDShaderScalarTransformer::min(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_min, dst, src1, src2);
}

void VDShaderScalarTransformer::max(int dst, int src1, int src2) {
	push_insn(kVDShaderInterpOp_max, dst, src1, src2);
}

void VDShaderScalarTransformer::rsq(int dst, int src) {
	push_insn(kVDShaderInterpOp_rsq, dst, src, 0);
}

void VDShaderScalarTransformer::rcp(int dst, int src) {
	push_insn(kVDShaderInterpOp_rcp, dst, src, 0);
}

void VDShaderScalarTransformer::frc(int dst, int src) {
	push_insn(kVDShaderInterpOp_frc, dst, src, 0);
}

void VDShaderScalarTransformer::log(int dst, int src) {
	push_insn(kVDShaderInterpOp_log, dst, src, 0);
}

void VDShaderScalarTransformer::exp(int dst, int src) {
	push_insn(kVDShaderInterpOp_exp, dst, src, 0);
}

void VDShaderScalarTransformer::neg(int dst, int src) {
	push_insn(kVDShaderInterpOp_neg, dst, src, 0);
}

void VDShaderScalarTransformer::abs(int dst, int src) {
	push_insn(kVDShaderInterpOp_abs, dst, src, 0);
}

void VDShaderScalarTransformer::cmp(int dst, int src1, int src2, int src3) {
	push_insn(kVDShaderInterpOp_cmp, dst, src1, src2, src3);
}

void VDShaderScalarTransformer::texld(int dst, int srcu, int srcv) {
	push_insn(kVDShaderInterpOp_texld, dst, srcu, srcv);
}

void VDShaderScalarTransformer::unpackr(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackr, dst, src, 0);
}

void VDShaderScalarTransformer::unpackg(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackg, dst, src, 0);
}

void VDShaderScalarTransformer::unpackb(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpackb, dst, src, 0);
}

void VDShaderScalarTransformer::unpacka(int dst, int src) {
	push_insn(kVDShaderInterpOp_unpacka, dst, src, 0);
}

void VDShaderScalarTransformer::beginloop() {
	push_insn(kVDShaderInterpOp_beginloop, 0, 0, 0);
}

void VDShaderScalarTransformer::push_insn(int op, int dst, int src1, int src2, int src3) {
	VDShaderInterpreterTransformedInsn tinsn;
	tinsn.op = op;
	tinsn.dst = dst;
	tinsn.src1 = src1;
	tinsn.src2 = src2;
	tinsn.src3 = src3;
	mShader.mInsns.push_back(tinsn);
}
